//NOTE: This is an autogenerated file - do not modify

struct TogAmbientLight
{
	float4	ambient;
};

struct TogDirectionalLight
{
	float3 ambient;
	float3 diffuse;
	float3 specular;
	float4 position;
};

struct TogPointLight
{
	float3 ambient;
	float3 diffuse;
	float3 specular;
	float4 position;
	float3 attenuation;
};

struct TogSpotLight
{
	float3 ambient;
	float3 diffuse;
	float3 specular;
	float3 direction;
	float4 position;
	float3 expCutOff;
	float3 attenuation;
};

struct TogMaterial
{
	float3 ambient;
	float4 diffuse;
	float3 specular;
	float3 emission;
	int    shininess;
};

struct TogTexGen
{
	float4 plane;
	float  enabled;
	float  mode;
};

float attenuate(float3 a,float3 b,float3 atten)
{
	float d = distance(a,b);
	return 1.0f / (atten.x + atten.y*d + atten.z*d*d);
}

float spotEffect(float3 L,float3 D,float cutOff,int exponent)
{
	float ang = max(dot(normalize(L),normalize(D)),0.0f);
	if(ang > cutOff)
		return pow(ang,exponent);
	else
		return 0.0f;
}

float dualConeSpotEffect(float3 L,float3 D,float cutOff,float outerCutOff,int exponent)
{
	float ang = max(dot(normalize(L),normalize(D)),0.0f);
	return pow(smoothstep(outerCutOff,cutOff,ang), exponent);
}
float2 modf(float2 a)
{
	return a-floor(a);
}
float linstep(float min,float max,float v)
{
	return clamp((v-min)/(max-min),0,1);
}
float ReduceLightBleeding(float p_max,float amount)
{
	return linstep(amount,1,p_max);
}
float ChebyshevUpperBound(float2 moments,float depth)
{
	float p = (depth <= moments.x);
	
	float variance = moments.y - (moments.x*moments.x);
	variance = max(variance,0.0) + 0.01;
	//variance = min(variance,1.0);
	
	float mean = moments.x-depth;
	float p_max = variance / (variance + mean*mean);
	float l = max(p,p_max);
	
	return smoothstep(0.01,1.0,l);
}
/////////////////////////////////////////////////////////////
//
// transfer functions to/from linear space
//

// Note - Step returns only 0 or 1 - so this is using lerp as an if/else block across all float4 colour values simultaneously

float gamma_PQ_To_Linear(float value, float diffuseWhite)
{
	float ePow = pow(saturate(value), 0.0126833);
	float outColour = pow((max(ePow - 0.8359375, 0.0f)) / ( 18.8515625 - 18.6875*ePow ), 6.277394);
	
	outColour /= diffuseWhite;
	return outColour;
}

float4 gamma_PQ_To_Linear(float4 inColour, float diffuseWhite)
{
	return float4(	gamma_PQ_To_Linear(inColour.x, diffuseWhite),
					gamma_PQ_To_Linear(inColour.y, diffuseWhite),
					gamma_PQ_To_Linear(inColour.z, diffuseWhite),
					inColour.w);
}

float gamma_HLG_To_Linear(float value, float diffuseWhite)
{
	float outValue;

	if(value < 0.0f)
		outValue = -(value * value / 3.0f);	
	else if(value <= 0.5f)
		// saturate is necessary - we use --fastmath which means powf() with a negative base returns nan.
		outValue = value * value / 3.0f;
	else
		outValue = (exp((value - 0.55991073f) / 0.17883277f) + 0.28466892f) / 12.0f;

	//if(outValue < 0.0)
	//	outValue = -pow(-outValue, 1.2f);
	//else
	//	outValue = pow(outValue, 1.2f);

	outValue /= diffuseWhite;
	return outValue;
}

float4 gamma_HLG_To_Linear(float4 inColour, float diffuseWhite)
{
	return float4(	gamma_HLG_To_Linear(inColour.x, diffuseWhite),
					gamma_HLG_To_Linear(inColour.y, diffuseWhite),
					gamma_HLG_To_Linear(inColour.z, diffuseWhite),
					inColour.w);
}

float gamma_Linear_To_HLG(float value, float diffuseWhite)
{
	float outValue;

	outValue = value * diffuseWhite;

	//if(outValue < 0.0)
	//	outValue = -pow(-outValue, 1.0f/1.2f);
	//else
	//	outValue = pow(outValue, 1.0f/1.2f);

	if(outValue >= 0.0833333f)
		outValue = (0.17883277f * log(12.0f * outValue - 0.28466892f) +  0.55991073f);
	else if(outValue >= 0.000000f)
		outValue = sqrt(3.0f*outValue);
	else
		outValue = -sqrt(3.0f*-outValue);

	return outValue;
}

float4 gamma_Linear_To_HLG(float4 inColour, float diffuseWhite)
{
	return float4(
		gamma_Linear_To_HLG(inColour.x, diffuseWhite),
		gamma_Linear_To_HLG(inColour.y, diffuseWhite),
		gamma_Linear_To_HLG(inColour.z, diffuseWhite),
		inColour.a);

}

float4 gamma_Rec709_To_Linear(float4 inColour)
{
	return float4(lerp(	inColour/4.5f, 
					pow((saturate(inColour) + 0.099f) / 1.099f, 1.0f/0.45f), 
					step(float4(0.081f), inColour)).rgb, 
					inColour.a);
}

float4 gamma_Linear_To_Rec709(float4 inColour)
{
	return float4(lerp(	4.5f * inColour, 
					1.099f * pow(saturate(inColour), 0.45f) - 0.099f, 
					step(float4(0.018f), inColour)).rgb,
					inColour.a);
}

float4 gamma_sRgb_To_Linear(float4 inColour)
{
	return float4(lerp(	inColour/12.92f, 
					pow((saturate(inColour) + 0.055f) / 1.055f, 2.4f), 
					step(float4(0.04045f), inColour)).rgb,
					inColour.a);
}

float gamma_Linear_To_sRgb(float value)
{
	if(value >= 0.0031308f)
		return pow(1.055f * value, 1.0f/2.4f) - 0.055f;

	return 12.92f * value;
}

float4 gamma_Linear_To_sRgb(float4 inColour)
{
	return float4(
		gamma_Linear_To_sRgb(inColour.x),
		gamma_Linear_To_sRgb(inColour.y),
		gamma_Linear_To_sRgb(inColour.z),
		inColour.w);
}

//
// Applies zebra affect to colours outside of the 0-1 range
// texcoord is used to apply the striping effect
//
// @param zebra   This float4 contains the following values :
//					.x : 1 to display zebra, if 0, no zebra is displayed.
//					.y : 1 to display zebra pattern for negative (invalid) RGB values.
//					.z : the level above which zebra will be displayed.
//					.w : an animation value to animate the zebra pattern
float4 gamma_applyZebra(float2 texCoord, float4 inColour, float4 zebra)
{
	float4 tooHigh = min(zebra.xxxx, step(zebra.zzzz, inColour));
	float4 tooLow = min(zebra.yyyy, step(inColour, float4(-0.0001f)));

	float4 outColour = inColour;
	outColour = lerp(	outColour, 
						lerp(	outColour, 
								float4(0.0f), 
								step(	10.0f, 
										fmod(texCoord.x+texCoord.y+(zebra.w*20.0f), 20.0f)
								)
							), 
						tooHigh);
	outColour  = lerp(	outColour, 
						lerp(	outColour, 
								float4(1.0f, 0.0f, 1.0f, 1.0f), 
								step(	10.0f, 
										fmod(texCoord.x+texCoord.y+(zebra.w*20.0f)+5, 20.0f)
									)
							), 
						tooLow);

	return float4(outColour.xyz, 1.0f);
}


uniform int lightingEnabled < string BindAddress = "Tog.Lighting.Enabled"; >;

uniform float4  clip04Enabled < string BindAddress = "Tog.ClipPlane04Enabled"; >;
uniform float4  clip04SoftDistance < string BindAddress = "Tog.ClipPlane04SoftDistance"; >;



//texture samplers

//application to vertex structure
struct VertIn
{
	float4 position : POSITION;
	float3 normal : NORMAL;
	float4 colour : COLOR;
};

//vertex to fragment structure
struct VertOut
{
	float4 position		:POSITION;
	float3 normal		:TEXCOORD0;
	float4 eyePos		:TEXCOORD1;
	float4 clipValues	:TEXCOORD2;
	float4 colour		:COLOR;
};


//the vertex program
VertOut vertex(
	VertIn IN
)
{
	VertOut OUT;

	OUT.position = mul(glstate.matrix.mvp, IN.position);
	OUT.eyePos = mul(glstate.matrix.modelview[0], IN.position);
	OUT.normal = normalize(mul((float3x3)glstate.matrix.invtrans.modelview[0], IN.normal));
	OUT.colour = IN.colour;
	OUT.clipValues = 0.0;
	if(clip04Enabled.x == 1.0f)
		OUT.clipValues.x = dot(OUT.eyePos.xyz,glstate.clip[0].plane.xyz) + glstate.clip[0].plane.w;
	if(clip04Enabled.y == 1.0f)
		OUT.clipValues.y = dot(OUT.eyePos.xyz,glstate.clip[1].plane.xyz) + glstate.clip[1].plane.w;
	if(clip04Enabled.z == 1.0f)
		OUT.clipValues.z = dot(OUT.eyePos.xyz,glstate.clip[2].plane.xyz) + glstate.clip[2].plane.w;
	if(clip04Enabled.w == 1.0f)
		OUT.clipValues.w = dot(OUT.eyePos.xyz,glstate.clip[3].plane.xyz) + glstate.clip[3].plane.w;




	return OUT;
}

//the pixel program
float4 pixel(
	VertOut IN
) : COLOR
{
	//the combined texture colours
	float4 decalColour = {1.0f,1.0f,1.0f,1.0f};
	//the overall diffuse and ambient lighting terms
	float3 lightColour = {0.0f,0.0f,0.0f};
	//the overall specular lighting term
	float3 lightSpecular = {0.0f,0.0f,0.0f};
	//temporary helper variables
	float3 ambientC,diffuseC,specularC;

	//clip?
	float clipAlpha = 1.0f;
	if(any(clip04Enabled))
	{
		//soft clipping
		if(any(clip04SoftDistance))
		{
			IN.clipValues += max(-clip04SoftDistance,0);
			float4 clips = (IN.clipValues + (1-clip04Enabled)) / (abs(clip04SoftDistance) + (1-clip04Enabled));
			clipAlpha = clamp(min(min(min(clips.x,clips.y),clips.z),clips.w),0,1);
		}
		clip(IN.clipValues);
	}


	//normal and eye vector
	float3 N = normalize(IN.normal);
	float3 E = normalize(IN.eyePos.xyz);

	float4 decalModulate;
	if(lightingEnabled > 0.9)
		decalModulate = float4(glstate.material.front.ambient.rgb*glstate.lightmodel.ambient.rgb + glstate.material.front.emission.rgb + lightColour, 1.0);
	else
		decalModulate = IN.colour;

	float4 decalTemp;

	if(lightingEnabled > 0.9)
		decalModulate.a *= glstate.material.front.diffuse.a;
	decalModulate.a *= clipAlpha;
	return float4( decalModulate + lightSpecular, decalModulate.a);
}

//vertex to fragment structure
technique generated
{
	pass p0
	{
		VertexProgram = compile gp4vp vertex();
		FragmentProgram = compile gp4fp pixel();
	}
}


